#include <cstring>
#include <iostream>
#include <fstream>
#include <cmath>
#include <iomanip>
#include <sys/time.h>
#include <stdlib.h>
#include <algorithm>

#define DEB(A)

using namespace std ;

struct task {
	int C ;
	double U ;
	double avg_lat, max_lat ;
} ;

// SHARED GLOBAL VARIABLES!!!!!

bool test_mode = false ; // if true, the results are checked against
		        // those of the brute-force algorithm
int M ; // current number of processors
task *ts ; // current task set
int N ; // current number of tasks
int ceil_U_sum ; // ceil(U_sum)
static enum {GAMMA, OMEGA} target_coeff ; // toggle to control the value
				   // computed by computeGammaOmega()
double Gamma, Omega ; // containers in which Gamma and Omega are
		      // stored after when ready,

double result ; // variable containing the value computed by
		// computeGammaOmega(): if this function is executed
		// after setting target_coeff to GAMMA or OMEGA, then
		// the function stores in result a value from which
		// Gamma or Omega can be immediately or almost
		// immediately computed

double max_result ; // variable containing the value from which Omega is
		  // immediately computed in calc_base_th_bound_BB

int *C_ordered_idxes, *U_ordered_idxes ;
int *chosen_tasks; // sequence of the indexes of the tasks already
                   // chosen for the current recursive step
bool *task_is_chosen ; // set of the indexes of the tasks already
		       // chosen for the current recursive step: the
		       // i-th element is true only if the i-th task
		       // is currently chosen

// sequences of U and C (indexes of the
// tasks with these U and these C)
// corresponding to the current upper bound
// to the harmonic bound. Before recursion
// starts, this is a true upper bound;
// then, at any recursive step, the first
// parts of these sequences are used to
// obtain an upper bound to the values that
// can be reached given the current tail
// sub-sequence of tasks already chosen
// (see previous variables). We can afford
// to keep a single, global copy of these
// sequences, because they are always reset
// before being used.
int *free_C_inc, *free_U_dec;

// minimum possible sum of U used to compute the current upper bound
// to the harmonic bound
double min_denominator_sum;

int num_tasks ; // number of tasks to consider in the formulas for Gamma or Omega

// prototype of function used in brute-force algorithm
double calc_lambda_min(const task in_ts[], int num_in_ts, double M_g,
		       int num_summed) ;


enum sorting_key_t {C, U, CU} ;

/*
 * create an index vector of num_indexes elements (indexes), with
 * indexes pointing to the tasks with the highest values for the key,
 * and sorted in increasing order of the key. If rm_idx is larger than
 * -1, then the index rm_idx is not inserted into the vector.
 */
void create_indexes(int *&indexes, int num_indexes, sorting_key_t key,
		    int rm_idx)
{
	DEB(cout<<"Reordering for key "<<key<<endl) ;

	if ((rm_idx <= -1 && N < num_indexes) ||
	    (rm_idx > -1 && N < num_indexes - 1) ) {
		cout<<"Too few tasks to index: "
		    <<N<<" "<<num_indexes<<" "<<rm_idx
		    <<endl ;
		exit(-1);
	}

	int indexes_whole_size = N ;
	if (rm_idx != -1)
		indexes_whole_size-- ;

	indexes = new int[indexes_whole_size] ;

	// init vector with increasing indexes
	for (int i = 0, j = 0 ; i < N ; i++) {
		if (i == rm_idx)
			continue ;
		indexes[j] = i ;
		j++ ;
	}

	// resort vector in decreasing order of the key of the indexed
	// tasks
	for (int i = 0 ; i < num_indexes ; i++) {
		int max_idx = i ;
		for (int j = i + 1 ; j < indexes_whole_size ; j++) {
			task max_task = ts[indexes[max_idx]] ;
			task cur_task = ts[indexes[j]] ;
			bool smaller ;

			switch(key) {
			case C:
				smaller = max_task.C < cur_task.C ;
				break ;
			case U:
				smaller = max_task.U < cur_task.U ;
				break ;
			case CU:
				smaller = max_task.C * max_task.U <
						cur_task.C * cur_task.U ;
				break ;
			}
			if (smaller)
				max_idx = j ;
		}
		if (max_idx != i) {
			int tmp = indexes[i] ;
			indexes[i] = indexes[max_idx] ;
			indexes[max_idx] = tmp ;
		}
	}

        // reverse order of first num_indexes elements
	int half = num_indexes / 2 ;

	for (int i = 0 ; i < half ; i++) {
		int tmp = indexes[i] ;
		indexes[i] = indexes[num_indexes - 1 - i] ;
		indexes[num_indexes - 1 - i] = tmp ;
	}

	// check that order is correct; REMOVE ONCE DEBUGGED!
	for (int i = 0 ; i < num_indexes - 1 ; i++) {
		task cur_task = ts[indexes[i]] ;
		task next_task = ts[indexes[i+1]] ;
		bool smaller ;

		switch(key) {
		case C:
			smaller = next_task.C < cur_task.C ;
			break ;
		case U:
			smaller = next_task.U < cur_task.U ;
			break ;
		case CU:
			smaller = next_task.C * next_task.U <
					cur_task.C * cur_task.U ;
			break ;
		}
		if (smaller) {
			cout<<"Wrong order!"<<endl ;
			exit(-1) ;
		}
	}
	DEB(
		for (int i = 0 ; i < num_indexes ; i++) {
			task cur_task = ts[indexes[i]] ;

			switch(key) {
			case C:
				cout<<cur_task.C<<" " ;
				break ;
			case U:
				cout<<cur_task.U<<" " ;
				break ;
			case CU:
				cout<<cur_task.U*cur_task.C<<" " ;
				break ;
			}
		}
		cout<<endl ;
		);

}

void recompute_free_C_inc_U(int n)
{
	for (int i = 0, lastU = N - 1 ; i < n ; i++, lastU--) {
		// find the highest-U task not yet chosen
		while (task_is_chosen[U_ordered_idxes[lastU]])
			lastU-- ;
		free_U_dec[i] = U_ordered_idxes[lastU] ;
	}

	min_denominator_sum = 0;
	for (int i = 0, lastU = 0 ; i < n ; i++, lastU++) {
		// find the highest-U task not yet chosen
		while (task_is_chosen[U_ordered_idxes[lastU]])
			lastU++ ;
		min_denominator_sum += ts[U_ordered_idxes[lastU]].U ;
	}

	DEB(cout<<"New free_C_inc:") ;
	for (int i = n - 1, lastC = N - 1 ; i >= 0 ; i--, lastC--) {
		// find the highest-C task not yet chosen
		while (task_is_chosen[C_ordered_idxes[lastC]])
			lastC-- ;
		free_C_inc[i] = C_ordered_idxes[lastC] ;
		DEB(cout<<" "<<free_C_inc[i]) ;
	}
	DEB(cout<<endl) ;
}

double compute_resultUB(int p)
{
	double resultUB = 0 ;
	/*
	 * Compute fixed and free parts, separately, for the sum of
	 * the U_i in the denominator for Gamma. The reason is that we
	 * reuse the fixed part, for computing the upper bound to
	 * M^{pi}_{G+1}
	 */
	double fixed_denominator_sum = 0, free_denominator_sum = 0,
		denominator_sum ;

	DEB(cout<<"Free part:") ;
	for (int i = 0 ; i < p ; i++) {
		resultUB += ts[free_C_inc[i]].C / (M - free_denominator_sum) ;

		DEB(cout<<" "
		    <<ts[free_C_inc[i]].C<<"/"<<(M - free_denominator_sum)) ;
		free_denominator_sum += ts[free_U_dec[i]].U ;
	}

	DEB(cout<<"\nFixed part:") ;
	for (int i = p ; i < num_tasks ; i++) {
		resultUB += ts[chosen_tasks[i]].C /
			(M - free_denominator_sum - fixed_denominator_sum) ;
		DEB(cout<<" "<<ts[chosen_tasks[i]].C
		    <<"/"
		    <<(M - free_denominator_sum - fixed_denominator_sum)) ;
		fixed_denominator_sum += ts[chosen_tasks[i]].U ;
	}
	DEB(cout<<endl) ;

	denominator_sum = free_denominator_sum + fixed_denominator_sum;

	if (target_coeff == OMEGA) {
		// add remaining parts to resultUB
		double second_sum = 0 ;
		denominator_sum = 0 ;
		DEB(cout<<"Omega extra free part:") ;
		for (int i = 0 ; i < p ; i++) {
			double next_denominator_sum = denominator_sum +
				ts[free_U_dec[i]].U ;
			second_sum += ts[free_U_dec[p - 1 - i]].U /
				((M - denominator_sum) *
				 (M - next_denominator_sum));

			DEB(cout<<" "<<ts[free_U_dec[i]].U
			    <<"/"
			    <<"("<<(M - denominator_sum)<<"*"
			    <<(M - next_denominator_sum)<<")") ;
			denominator_sum = next_denominator_sum ;
		}

		DEB(cout<<"\nOmega extra fixed part:") ;
		for (int i = p ; i < num_tasks ; i++) {
			double next_denominator_sum = denominator_sum +
				ts[chosen_tasks[i]].U ;
			second_sum += ts[chosen_tasks[i]].U /
				((M - denominator_sum) *
				 (M - next_denominator_sum));

			DEB(cout<<" "<<ts[chosen_tasks[i]].U
			    <<"/"
			    <<"("<<(M - denominator_sum)<<"*"
			    <<(M - next_denominator_sum)<<")") ;
			denominator_sum = next_denominator_sum ;
		}
		DEB(cout<<endl) ;

		DEB(cout<<"Second sum: "<<second_sum<<endl) ;
		resultUB += second_sum * Gamma ;
		resultUB *= M - (min_denominator_sum + fixed_denominator_sum);
		DEB(cout<<"ResultUB/M "<<resultUB/M<<endl) ;
	} else
		DEB(cout<<"First sum UB * M: "<<resultUB * M<<endl) ;

	return resultUB ;
}

inline int max ( int a, int b ) { return a > b ? a : b; }
inline double min_double ( double a, double b ) { return a < b ? a : b; }

void rec_computeGammaOmega(int p)
{
	DEB(cout<<"[rec_computeGamma] p="<<p<<endl) ;

#if 1
	// used to exclude some tasks from the search
	// double U_max = 3 ;
	double U_min = 3 ; // purposedly invalid value
#endif

#if 0 // does not seem correct
	if (p < num_tasks - 1 && p > 0) {
		U_max = 0 ;

		for (int i = N - 1 ; i >= 0; i--) {
			int new_chosen = C_ordered_idxes[i] ;

			// nothing to do if this task is already chosen
			if (task_is_chosen[new_chosen]) {
				DEB(cout<<"Task "<<new_chosen
				    <<" already picked, skipping"
				    <<endl) ;
				continue ;
			}
			if (U_max < ts[new_chosen].U)
				U_max = ts[new_chosen].U;
		}
	}
	bool waiting_choice = p < num_tasks - 1 && p > 0 ;
#endif

	// try all tasks not yet chosen for higher positions
	for (int i = N - 1 ; i >= 0; i--) {
		// in the first iteration, try to get the task with
		// the best C_i for position p. Then, in the second
		// iteration, the second best, and so on. The %N
		// guarantees that all tasks are scanned inside
		// C_ordered_idxes as i decreases.
		int new_chosen = C_ordered_idxes[i] ;

		// nothing to do if this task is already chosen
		if (task_is_chosen[new_chosen]) {
			DEB(cout<<"Task "<<new_chosen
			    <<" already picked, skipping"
			    <<endl) ;
			continue ;
		}
		DEB(cout<<"Checking"<<new_chosen<<" in position "<<p
		    <<" , picked in odered at idx "<<i
		    <<endl) ;

#if 1
		// new_chose is not convenient if, in addition to
		// having a lower C, it has also a higher U than the
		// tasks with a higher C
		if (U_min < 2 && p == num_tasks - 1 &&
		    ts[new_chosen].U >= U_min) {
			// fails for the first task considered
			DEB(cout<<"Task "<<new_chosen
			    <<" non convenient ("
			    <<ts[new_chosen].U<<" "<<U_min
			    <<"), skipping"
			    <<endl) ;

			U_min = min(ts[new_chosen].U, U_min) ;

			continue ; // no point in choosing this task:
				   // it is more convenient to have it
				   // in lower position
		} else
			U_min = min_double(ts[new_chosen].U, U_min) ;
#endif

#if 0 // does not seem correct
		if (p < num_tasks - 1 && p > 0) {
			if (waiting_choice)
				waiting_choice = false ;
			else if (ts[new_chosen].U >= U_max) {
				DEB(cout<<"Intermediate task "<<new_chosen
				    <<" non convenient ("
				    <<ts[new_chosen].U<<" "<<U_max
				    <<"), skipping"
				    <<endl) ;

				continue ;
			}
		}
#endif

		DEB(cout<<"Accepted task "<<new_chosen<<" in position "<<p
		    <<endl) ;
		task_is_chosen[new_chosen] = true ;

		// set p-th element in chosen_tasks to new_chosen
		chosen_tasks[p] = new_chosen ;

		// compute new values for chosen_C and chosen_U, for
		// the free part of the formula, and as a function of
		// the new choice for the task in position p
		recompute_free_C_inc_U(p) ;

		double resultUB = compute_resultUB(p) ;

		if (resultUB < result) // current result cannot be
				     // overcome with these chosen
				     // tasks
			goto end ;

		if (p == 0) {
			if (resultUB > result)
				result = resultUB ;
			DEB(cout<<"New result/M "<<result/M<<endl) ;
			goto end ;
		}

		rec_computeGammaOmega(p-1) ;

	end:
		// calculation finished with this task
		task_is_chosen[new_chosen] = false ;
	}
}

double computeRefValue(int ordered_idxes[])
{
	// sums used for computing a first value for result
	double sums[num_tasks + 1] ;

	// we exploit chosen_tasks too, to not create another array
	// (this array will be then progressively rewritten during
	// recursion)

	sums[0] = 0 ;

	// use the indexes passed as input to choose the other tasks
	for (int i = 0 ; i < num_tasks; i++)
		chosen_tasks[i] = ordered_idxes[i + N - num_tasks] ;

	DEB(cout<<"Chosen tasks (reference value):" ;
		for (int i = 0 ; i < num_tasks ; i++)
			cout<<" "<<chosen_tasks[i] ;
	    cout<<endl ;) ;

	DEB(cout<<"sums:") ;
	for (int i = 1 ; i < num_tasks + 1 ; i++) {
		sums[i] = sums[i-1] + ts[chosen_tasks[i-1]].U ;
		DEB(cout<<" "<<sums[i-1]<<"+"<<ts[chosen_tasks[i-1]].U
		    <<"="<<sums[i]) ;
	}
	DEB(cout<<endl) ;

	double private_result = 0 ;
	DEB(cout<<"First sum computation (reference value):") ;
	for (int i = 0 ; i < num_tasks ; i++) {
		private_result += ts[chosen_tasks[i]].C / (M - sums[i]) ;
		DEB(cout<<" "<<ts[chosen_tasks[i]].C<<"/"
		    <<(M - sums[i])) ;
	}
	DEB(cout<<endl) ;

	if (target_coeff == OMEGA) {
		// add remaining parts to result
		double second_sum = 0 ;
		for (int i = 0 ; i < num_tasks ; i++)
			second_sum += ts[chosen_tasks[i]].U /
				((M - sums[i]) * (M - sums[i+1]));
		private_result += second_sum * Gamma ;
		private_result *= M - sums[num_tasks] ;
		DEB(cout<<"Omega/M (reference value): "
		    <<private_result<<endl) ;
	} else
		DEB(cout<<"First sum (reference value): "
		    <<private_result<<endl) ;

	return private_result ;
}

void computeGammaOmega()
{
	DEB(cout<<"Target coefficient "<<target_coeff
	    <<", num_tasks "<<num_tasks
	    <<endl) ;

	double GammaBF ; // Gamma computed with brute force, for
			 // debugging

	// compute first tentative values for result
	DEB(cout<<"Compute reference value for C-ordered tasks"
	    <<endl) ;
	double private_result_C = computeRefValue(C_ordered_idxes);
	DEB(cout<<"Compute reference value for U-ordered tasks"
	    <<endl) ;
	double private_result_U = computeRefValue(U_ordered_idxes);

	double private_result = max(private_result_C, private_result_U);

	if (private_result > result) {
		DEB(cout<<"Increasing result from "
		    <<result<<" to "<<private_result<<endl) ;
		result = private_result ;
	}
	// reinit set of chosen tasks before computing resultUB: no
	// task chosen yet
	for (int i = 0 ; i < N ; i++)
		task_is_chosen[i] = false ;

	// compute an upper bound to result; to this purpose,
	// reset and reuse free_U_dec and free_C_inc
	recompute_free_C_inc_U(num_tasks) ;

	double resultUB = compute_resultUB(num_tasks) ;

	if (resultUB < result - 1e-9)
		return ;

	if (test_mode && target_coeff == GAMMA) {
		DEB(cout<<"Invoking calc_lambda_min ..."<<endl) ;
		GammaBF = calc_lambda_min(ts, N, M, 0) ;

		if (resultUB < GammaBF - 1e-9) {
			cout<<"GammaUB "<<resultUB*M
			    <<" lower than GammaBF "<<GammaBF*M
			    <<endl ;
			exit(1) ;
		}
		if (result > GammaBF + 1e-9) {
			cout<<"GammaBB "<<result*M
			    <<" higher than GammaBF "<<GammaBF*M
			    <<endl ;
			exit(1) ;
		}

	}

	if (target_coeff == OMEGA)
		DEB(cout<<"Omega before recursion: "
		    <<result/M<<" "<<resultUB/M<<endl) ;

	if (abs(result - resultUB) < 1e-9)
		DEB(cout<<"result "<<result<<" found at first step!"
		    <<endl) ;
	else {
		DEB(cout<<"Starting recursion ..."<<endl) ;

		rec_computeGammaOmega(num_tasks - 1) ;
	}

	if (test_mode  && target_coeff == GAMMA) {
		if (abs(result - GammaBF) > 1e-9) {
			cout<<"Gamma results do not match!"<<endl ;
			cout<<"BB "<<result*M<<", BF "<<GammaBF*M<<endl ;
			exit(1) ;
		} else
			DEB(cout<<"Gamma results coincide: "<<result*M
			    <<endl) ;
	}
}

double calc_base_th_bound_BB(const task ts[], int N)
{
	target_coeff = GAMMA ;
	num_tasks = ceil_U_sum - 1 ;
	result = 0 ;
	computeGammaOmega() ;
	Gamma = M * result ;

	target_coeff = OMEGA ;
	result = max_result = 0 ;
	DEB(cout<<"num_tasks "<<num_tasks<<" ceil_U_sum "
	    <<ceil_U_sum<<endl) ;
	for (num_tasks = ceil_U_sum - 1 ; num_tasks > 0 ; num_tasks--) {
		DEB(cout<<endl) ;
		computeGammaOmega() ;
		DEB(cout<<"Omega result/M for num_tasks == "<<num_tasks
		    <<" -> "<<result/M) ;

		if (result > max_result) {
			if (max_result > 0)
				cout<<" >>> Increasing max_result"
				    <<endl ;
			max_result = result ;
		}
		else
			result = max_result ; // increase result
					      // before next iteration

		DEB(cout<<"Omega max_result/M for num_tasks == "
		    <<num_tasks
		    <<" -> "<<max_result/M<<endl) ;
	}
	DEB(cout<<"max_result/M "<<max_result/M<<endl) ;

	Omega = max_result / M ;

	return Omega ;
}

// brute-force computation of Gamma (called lambda_min in first
// version of the paper)
double Lambda_min ; // current lambda_min
double calc_lambda_min(const task in_ts[], int num_in_ts, double M_g,
		       int num_summed) {
	if (num_in_ts == 0 || num_summed == ceil_U_sum - 1)
		return 0 ;

	double max_lambda_min = 0 ;

	task new_in_ts[num_in_ts - 1] ;

	for (int i = 0 ; i < num_in_ts ; i++) {
		if (i > 0)
			/*
			 * Before the next assignment, new_in_ts[i-1]
			 * contains in_ts[i]. With the next assignment,
			 * in_ts[i] is removed from new_in_ts
			 * (it will be inserted in next iteration)
			 */
			new_in_ts[i-1] = in_ts[i-1] ;
		else // generate new subset by removing first task
		    for (int j = 0 ; j < num_in_ts - 1 ; j++)
			new_in_ts[j] = in_ts[j + 1] ;

		DEB(for (int k = 0 ; k < num_summed ; k++)
			    cout<<"\t") ;

		DEB(cout<<"New task set: " ;
			for (int k = 0 ; k < num_in_ts - 1; k++)
			    cout<<new_in_ts[k].U<<" " ;
		    cout<<endl) ;

		double res = calc_lambda_min(new_in_ts, num_in_ts - 1,
				    M_g - in_ts[i].U, num_summed + 1) ;

		double lambda_min = in_ts[i].C/M_g + res ;

		DEB(for (int k = 0 ; k < num_summed ; k++)
			    cout<<"\t") ;

		DEB(cout<<"num_summed "<<num_summed
		    <<", i "<<i
		    <<", C_i "<<in_ts[i].C
		    <<", num_in_ts "<<num_in_ts<<", M_g "<<M_g
		    <<", lambda_min = "
		    <<in_ts[i].C<<"/"<<M_g<<"+"<<res
		    <<" = "<<lambda_min
		    <<endl );

		if (max_lambda_min < lambda_min)
			max_lambda_min = lambda_min ;
	}

	return max_lambda_min ;
}

double max_up_bound ; // shared variable used in next recursive
		      // function to return result!!!
int max_num_summed ; // shared variable used to control recursion

void _calc_sum_up_bound(const task in_ts[], int num_in_ts, double M_g,
		       int num_summed, double partial_sum)
{
    if (num_in_ts == 0 || num_summed == max_num_summed) {
	DEB(cout<<"M_M "<<M_g<<endl) ;

	double up_bound = M_g * partial_sum ;
	if(max_up_bound < up_bound) {
	    max_up_bound = up_bound ;
	    DEB(cout<<"update_max_up_bound "<<max_up_bound<<endl);
	}
	return ;
    }

    task new_in_ts[num_in_ts - 1] ;

    for (int i = 0 ; i < num_in_ts ; i++) {
	if (i > 0)
	    new_in_ts[i-1] = in_ts[i-1] ;
	else
	    for (int j = 0 ; j < num_in_ts - 1 ; j++)
		new_in_ts[j] = in_ts[j + 1] ;

	double new_partial_sum = partial_sum +
	    (Lambda_min * in_ts[i].U) / (M_g * (M_g - in_ts[i].U)) +
	    in_ts[i].C/M_g ;

	DEB(for (int k = 0 ; k < num_summed ; k++)
		cout<<"\t") ;

	DEB(cout<<"num_summed "<<num_summed
	    <<", i "<<i
	    <<", C_i "<<in_ts[i].C
	    <<", num_in_ts "<<num_in_ts<<", M_g "<<M_g
	    <<", partial_sum = "
	    <<partial_sum<<" + "
	    <<(Lambda_min * in_ts[i].U)<<"/"
	    <<"("<<M_g<<"*"<<(M_g - in_ts[i].U)<<")"
	    <<"+"
	    <<in_ts[i].C/M_g
	    <<" = "<<new_partial_sum
	    <<endl );

	_calc_sum_up_bound(new_in_ts, num_in_ts - 1,
			   M_g - in_ts[i].U,
			   num_summed + 1, new_partial_sum) ;
    }
}

void calc_sum_up_bound(const task in_ts[], int num_in_ts, double M_g,
		       int num_summed, double partial_sum)
{
	max_up_bound = 0 ;
	int num_summed_yielding_max = 0 ;

	for (max_num_summed = 0 ; max_num_summed <= ceil_U_sum - 1 ;
	     max_num_summed++) {
		DEB(cout<<"max_num_summed "<<max_num_summed<<endl) ;
		double prev_max_up_bound = max_up_bound ;
		_calc_sum_up_bound(in_ts, num_in_ts, M_g, num_summed, partial_sum) ;
		if (prev_max_up_bound > max_up_bound)
			max_up_bound = prev_max_up_bound ;
		else
			num_summed_yielding_max = max_num_summed ;
	}

	if (num_summed_yielding_max < ceil_U_sum - 1)
		cout<<"WARNING: max obtained with "<<num_summed_yielding_max
		    <<"/"<<ceil_U_sum-1<<endl ;
}

double calc_base_th_bound_BF(const task ts[], int N, int M)
{
	Lambda_min = M * calc_lambda_min(ts, N, M, 0) ;

	DEB(cout<<"Lambda_min: "<<Lambda_min<<endl) ;

	calc_sum_up_bound(ts, N, M, 0, 0) ;

	return max_up_bound / M ;
}

int main(int argc, char* argv[])
{
	if (argc < 4 || !strcmp(argv[1], "help") ) {
		cout<<"Usage: calc-bound_ratio-from-tasksets <input file> "
		    <<"<bounds outfile> <max-ratios outfile>"
		    <<"[<prof stats outfile> [max num of task sets to parse]]"
		    <<endl ;
		return 1 ;
	}

	ifstream f(argv[1]) ;

	if (! f) {
		cout<<"Failed to open input file "
		    <<argv[1]
		    <<", exiting."<<endl ;
		exit(-1) ;
	}
	ofstream bounds_file(argv[2]) ;
	ofstream ratios_file(argv[3]) ;

	int num_ts = 0, max_num_ts = 0 ;
#if 0 // statistics deactivated
	ofstream prof_file ;

	bool prof_mode = false ;

	double avg_N = 0, avg_M = 0 ;
	if (argc > 4) {
		prof_mode = true ;
		prof_file.open(argv[4]) ;
		if (! prof_file) {
			cout<<"Failed to open profile file "
			    <<argv[4]
			    <<", exiting."<<endl ;
			exit(-1) ;
		}
	}
#endif

	if (argc == 6)
		max_num_ts = atoi(argv[5]) ;

	int N_heavy, N_light ;

	struct timeval start, finish ;
	gettimeofday(&start, NULL);
	while(f>>M>>N_heavy>>N_light &&
		(
#if 0 // profiling deactivated

			!prof_mode ||
#endif
		 max_num_ts == 0 || num_ts < max_num_ts
			)) {
		num_ts++ ;
		N = N_heavy + N_light ;
#if 0 // profiling deactivated
		avg_N += N ;
		avg_M += M ;
#endif
		if(N > 10) // long per-task-set execution time
			DEB(cout<<"Working on ts "<<num_ts
			    <<", with N "<<N<<endl) ;

		double M_1M = ((double) (M - 1)) / M ;

		// cout<<M<<" "<<N<<endl ;

		ts = new task[N] ;

		double U_sum = 0 ;

		for (int i = 0 ; i < N ; i++) {
			double period ;

			f>>ts[i].C ;
			f>>period ;
			ts[i].U = ts[i].C / period ;
			DEB(cout<<"Task "<<i<<": "<<ts[i].C<<" "
			    <<ts[i].U<<endl );
			U_sum += ts[i].U ;
			DEB(cout<<"Task "<<i<<": "<<ts[i].C<<" "
			    <<ts[i].U
			    <<", partial U_sum "<<U_sum
			    <<endl );
		}
		ceil_U_sum = ceil(U_sum) ;

		if (ceil_U_sum > M) {
			if (ceil_U_sum > M + 1) {
				cout<<"ceil_U_sum "<<ceil_U_sum
				    <<" > M "<<M
				    <<endl ;
				exit(-1) ;
			} else
				ceil_U_sum = M ;

		}
		DEB(cout<<"ceil_U_sum "<<ceil_U_sum<<endl) ;

		create_indexes(C_ordered_idxes, N, C, -1) ;
		create_indexes(U_ordered_idxes, N, U, -1) ;
		task_is_chosen = new bool[N] ;
		chosen_tasks = new int[N] ;
		free_U_dec = new int[N] ;
		free_C_inc = new int[N] ;

		double base_th_bound_BB = calc_base_th_bound_BB(ts, N) ;

		if (test_mode) {
			// compute base bound with brute force
			double base_th_bound_BF =
				calc_base_th_bound_BF(ts, N, M) ;
			DEB(cout<<"Base th. bound BF: "
			    <<base_th_bound_BF<<endl) ;

			if (abs(base_th_bound_BF - base_th_bound_BB) > 1e-9) {
				cout<<"Fatal: BB "<<base_th_bound_BB
				    <<" and BF "<<base_th_bound_BF
				    <<" do not coincide!"
				    <<endl ;
				exit(-1) ;
			}
		}

		double max_ratio = 0 ;

		for (int i = 0 ; i < N ; i++) {
			f>>ts[i].avg_lat>>ts[i].max_lat ;

			if (!f) {
				cout<<"Input file truncated"<<endl ;
				return 1 ;
			}

			double th_bound = base_th_bound_BB + M_1M * ts[i].C ;

			bounds_file<<th_bound<<endl ;

			double ratio = ts[i].max_lat / th_bound ;

			if (ratio > max_ratio)
				max_ratio = ratio ;
		}

		ratios_file<<fixed<<max_ratio<<endl ;
		delete [] ts ;
		ts = 0 ;
		delete [] C_ordered_idxes ;
		delete [] U_ordered_idxes ;
		delete [] task_is_chosen ;
		delete [] chosen_tasks ;
		delete [] free_U_dec ;
		delete [] free_C_inc ;

		DEB(cout<<"Finished ts "<<num_ts
		    <<", with N "<<N<<endl) ;

	}
	gettimeofday(&finish, NULL);
	finish.tv_sec -= start.tv_sec;
	finish.tv_usec -= start.tv_usec;
	if (finish.tv_usec < 0) {
		finish.tv_usec += 1000000;
		finish.tv_sec--;
	}

	if (num_ts == 0) {
	    cout<<"Empty or corrupted input file!!!!"<<endl ;
	    return -1 ;
	}

#if 0 // statistics deactivated
	unsigned long long int tot_duration = finish.tv_sec*1000000 +
		finish.tv_usec;
    	double avg_duration = (double) tot_duration / num_ts ;

	avg_N /= num_ts ;
	avg_M /= num_ts ;

	prof_file<<"Total duration "<<tot_duration/1000<<"ms"
	    <<", num ts "<<num_ts
	    <<", avg duration "<<avg_duration<<"us"
	    <<", avg N "<<avg_N
	    <<", avg M "<<avg_M
	    <<endl ;
	cout<<"Total duration "<<tot_duration/1000<<"ms"
	    <<", num ts "<<num_ts
	    <<", avg duration "<<avg_duration<<"us"
	    <<", avg N "<<avg_N
	    <<", avg M "<<avg_M
	    <<endl ;
#endif

	return 0 ;
}
