#include <cstring>
#include <iostream>
#include <fstream>
#include <cmath>
#include <iomanip>
#include <sys/time.h>
#include <stdlib.h>

#define DEB(A)

using namespace std ;

struct task {
	int C ;
	double U ;
	double avg_lat, max_lat ;
} ;

// SHARED GLOBAL VARIABLES!!!!!

int M ; // current number of processors
int U_ceil ; // ceil(U_sum)
double Lambda_min ; // current lambda_min

double calc_lambda_min(const task in_ts[], int num_in_ts, double M_g,
		       int num_summed) {
	if (num_in_ts == 0 || num_summed == U_ceil - 1)
		return 0 ;

	double max_lambda_min = 0 ;
	
	task new_in_ts[num_in_ts - 1] ;

	for (int i = 0 ; i < num_in_ts ; i++) {
		if (i > 0)
			/*
			 * Before the next assignment, new_in_ts[i-1]
			 * contains in_ts[i]. With the next assignment,
			 * in_ts[i] is removed from new_in_ts
			 * (it will be inserted in next iteration)
			 */
			new_in_ts[i-1] = in_ts[i-1] ;
		else // generate new subset by removing first task
		    for (int j = 0 ; j < num_in_ts - 1 ; j++)
			new_in_ts[j] = in_ts[j + 1] ;
		
		DEB(for (int k = 0 ; k < num_summed ; k++)
			    cout<<"\t") ;

		DEB(cout<<"New task set: " ;
			for (int k = 0 ; k < num_in_ts - 1; k++)
			    cout<<new_in_ts[k].U<<" " ;
		    cout<<endl) ;
		
		double res = calc_lambda_min(new_in_ts, num_in_ts - 1,
				    M_g - in_ts[i].U, num_summed + 1) ;

		double lambda_min = in_ts[i].C/M_g + res ;
		
		DEB(for (int k = 0 ; k < num_summed ; k++)
			    cout<<"\t") ;

		DEB(cout<<"num_summed "<<num_summed
		    <<", i "<<i
		    <<", C_i "<<in_ts[i].C
		    <<", num_in_ts "<<num_in_ts<<", M_g "<<M_g
		    <<", lambda_min = "
		    <<in_ts[i].C<<"/"<<M_g<<"+"<<res
		    <<" = "<<lambda_min
		    <<endl );

		if (max_lambda_min < lambda_min)
			max_lambda_min = lambda_min ;
	}

	return max_lambda_min ;
}

double max_up_bound ; // shared variable used in next recursive
		      // function to return result!!!
int max_num_summed ; // shared variable used to control recursion

void _calc_sum_up_bound(const task in_ts[], int num_in_ts, double M_g,
		       int num_summed, double partial_sum)
{
    if (num_in_ts == 0 || num_summed == max_num_summed) {
	DEB(cout<<"M_M "<<M_g<<endl) ;

	double up_bound = M_g * partial_sum ;
	if(max_up_bound < up_bound) {
	    max_up_bound = up_bound ;
	    DEB(cout<<"update_max_up_bound "<<max_up_bound<<endl);
	}
	return ;
    }

    task new_in_ts[num_in_ts - 1] ;

    for (int i = 0 ; i < num_in_ts ; i++) {
	if (i > 0)
	    new_in_ts[i-1] = in_ts[i-1] ;
	else
	    for (int j = 0 ; j < num_in_ts - 1 ; j++)
		new_in_ts[j] = in_ts[j + 1] ;

	double new_partial_sum = partial_sum +
	    (Lambda_min * in_ts[i].U) / (M_g * (M_g - in_ts[i].U)) +
	    in_ts[i].C/M_g ;
		
	DEB(for (int k = 0 ; k < num_summed ; k++)
		cout<<"\t") ;

	DEB(cout<<"num_summed "<<num_summed
	    <<", i "<<i
	    <<", C_i "<<in_ts[i].C
	    <<", num_in_ts "<<num_in_ts<<", M_g "<<M_g
	    <<", partial_sum = "
	    <<partial_sum<<" + "
	    <<(Lambda_min * in_ts[i].U)<<"/"
	    <<"("<<M_g<<"*"<<(M_g - in_ts[i].U)<<")"
	    <<"+"
	    <<in_ts[i].C/M_g
	    <<" = "<<new_partial_sum
	    <<endl );

	_calc_sum_up_bound(new_in_ts, num_in_ts - 1,
			   M_g - in_ts[i].U,
			   num_summed + 1, new_partial_sum) ;
    }
}

void calc_sum_up_bound(const task in_ts[], int num_in_ts, double M_g,
		       int num_summed, double partial_sum)
{
	max_up_bound = 0 ;
	int num_summed_yielding_max = 0 ;

	for (max_num_summed = 0 ; max_num_summed <= U_ceil - 1 ; max_num_summed++) {
		DEB(cout<<"max_num_summed "<<max_num_summed<<endl) ;
		double prev_max_up_bound = max_up_bound ;
		_calc_sum_up_bound(in_ts, num_in_ts, M_g, num_summed, partial_sum) ;
		if (prev_max_up_bound > max_up_bound)
			max_up_bound = prev_max_up_bound ;
		else
			num_summed_yielding_max = max_num_summed ;
	}

	if (num_summed_yielding_max < U_ceil - 1)
		cout<<"WARNING: max obtained with "<<num_summed_yielding_max
		    <<"/"<<U_ceil-1<<endl ;
}

double calc_base_th_bound(const task ts[], int N, int M)
{
	Lambda_min = M * calc_lambda_min(ts, N, M, 0) ;

	DEB(cout<<"Lambda_min: "<<Lambda_min<<endl) ;
	
	calc_sum_up_bound(ts, N, M, 0, 0) ;
	
	return max_up_bound / M ;
}

int main(int argc, char* argv[])
{
	if (argc < 4 || !strcmp(argv[1], "help") ) {
		cout<<"Usage: calc-bound_ratio-from-tasksets <input file> "
		    <<"<bounds outfile> <max-ratios outfile>"
		    <<"[<prof stats outfile> [max num of task sets to parse]]"
		    <<endl ;
		return 1 ;
	}
	
	ifstream f(argv[1]) ;
	
	if (! f) {
		cout<<"Opening of input file "
		    <<argv[1]
		    <<" failed, exiting."<<endl ;
		exit(-1) ;
	}
	ofstream bounds_file(argv[2]) ;
	ofstream ratios_file(argv[3]) ;
	ofstream prof_file ;

	bool prof_mode = false ;
	int num_ts = 0, max_num_ts = 0 ;
	double avg_N = 0, avg_M = 0 ;

	if (argc > 4) {	
		prof_mode = true ;
		prof_file.open(argv[4]) ;
		if (! prof_file) {
			cout<<"Opening of profile file "
			    <<argv[4]
			    <<" failed, exiting."<<endl ;
			exit(-1) ;
		}
		if (argc == 6)
			max_num_ts = atoi(argv[5]) ;
	}

	int N_heavy, N_light ;

	struct timeval start, finish ;
	gettimeofday(&start, NULL);
	while(f>>M>>N_heavy>>N_light &&
		(!prof_mode || max_num_ts == 0 || num_ts < max_num_ts)) {
		num_ts++ ;
		int N = N_heavy + N_light ;
		avg_N += N ;
		avg_M += M ;

		if(N > 10) // long per-task-set execution time
			DEB(cout<<"Working on ts "<<num_ts
			    <<", with N "<<N<<endl) ;

		double M_1M = ((double) (M - 1)) / M ;

		// cout<<M<<" "<<N<<endl ;

		task ts[N] ;

		double U_sum = 0 ; 

		for (int i = 0 ; i < N ; i++) {
			double period ;

			f>>ts[i].C ;
			f>>period ;
			ts[i].U = ts[i].C / period ;
			DEB(cout<<"Task "<<i<<": "<<ts[i].C<<" "
			    <<ts[i].U<<endl );
			U_sum += ts[i].U ;
			DEB(cout<<"Task "<<i<<": "<<ts[i].C<<" "
			    <<ts[i].U
			    <<", partial U_sum "<<U_sum
			    <<endl );
		}
		U_ceil = ceil(U_sum) ;

		if (U_ceil > M) {
			if (U_ceil > M + 1) {
				cout<<"U_ceil "<<U_ceil
				    <<" > M "<<M
				    <<endl ;
				exit(-1) ;
			} else
				U_ceil = M ;
			
		}
		DEB(cout<<"U_ceil "<<U_ceil<<endl) ;
		
		double base_th_bound = calc_base_th_bound(ts, N, M) ;
		DEB(cout<<"Bas th. bound: "<<base_th_bound<<endl) ;
		  
		double max_ratio = 0 ;

		for (int i = 0 ; i < N ; i++) {
			f>>ts[i].avg_lat>>ts[i].max_lat ;

			if (!f) {
			    cerr<<"Input file truncated"<<endl ;
			    return 1 ;
			}

			double th_bound = base_th_bound + M_1M * ts[i].C ;
			
			bounds_file<<th_bound<<endl ;
				
			double ratio = ts[i].max_lat / th_bound ;
			
			if (ratio > max_ratio)
				max_ratio = ratio ;
		}
		
		ratios_file<<fixed<<max_ratio<<endl ;
	}
	gettimeofday(&finish, NULL);
	finish.tv_sec -= start.tv_sec;
	finish.tv_usec -= start.tv_usec;
	if (finish.tv_usec < 0) {
		finish.tv_usec += 1000000;
		finish.tv_sec--;
	}

	unsigned long long int tot_duration = finish.tv_sec*1000000 +
		finish.tv_usec;
    	double avg_duration = (double) tot_duration / num_ts ;

	avg_N /= num_ts ;
	avg_M /= num_ts ;

	prof_file<<"Total duration "<<tot_duration/1000<<"ms"
	    <<", num ts "<<num_ts
	    <<", avg duration "<<avg_duration<<"us"
	    <<", avg N "<<avg_N
	    <<", avg M "<<avg_M
	    <<endl ;
	cout<<"Total duration "<<tot_duration/1000<<"ms"
	    <<", num ts "<<num_ts
	    <<", avg duration "<<avg_duration<<"us"
	    <<", avg N "<<avg_N
	    <<", avg M "<<avg_M
	    <<endl ;

	return 0 ;
}
