#include <cstring>
#include <iostream>
#include <fstream>
#include <cmath>
#include <iomanip>
#include <limits>
#include <sys/time.h>
#include <stdlib.h>

#define DEB(A)

using namespace std ;

struct task {
	int C ;
	double P ;
	double avg_lat, max_lat ;
} ;

// SHARED GLOBAL VARIABLES!!!!!

int M ; // current number of processors
int U_ceil ; // ceil(U_sum)

int main(int argc, char* argv[])
{
	if (argc < 4 || !strcmp(argv[1], "help") ) {
		cout<<"Usage: calc-norm-error <simulation-results input file> "
		    <<"<lateness-bounds input file> <ratios outfile>"
		    <<endl ;
		return 1 ;
	}
	
	ifstream sim_res_file(argv[1]) ;
	
	if (! sim_res_file) {
		cout<<"Opening of input file "<<argv[1]
		    <<" failed, exiting."<<endl ;
		exit(-1) ;
	}

	ifstream lat_bounds_file(argv[2]) ;
	
	if (! lat_bounds_file) {
		cout<<"Opening of input file "<<argv[2]
		    <<" failed, exiting."<<endl ;
		exit(-1) ;
	}
	
	ofstream ratios_file(argv[3]) ;

	int num_ts = 0 ;
	int N_heavy, N_light ;

	while(sim_res_file>>M>>N_heavy>>N_light) {
		num_ts++ ;
		int N = N_heavy + N_light ;

		DEB(cout<<"Working on ts "<<num_ts
		    <<", with N "<<N<<endl) ;

		DEB(cout<<M<<" "<<N<<endl) ;

		if (N <= 0) {
			cout<<"Fatal error: N == "<<N<<endl ;
			exit(-1) ;
		}
		
		task ts[N] ;

		for (int i = 0 ; i < N ; i++) {
			sim_res_file>>ts[i].C>>ts[i].P ;
		}
		
		sim_res_file>>ts[0].avg_lat>>ts[0].max_lat ;
		double th_bound ;
		lat_bounds_file>>th_bound ;

		if (!lat_bounds_file) {
			lat_bounds_file.clear() ;
			char c ;
			lat_bounds_file>>c ;
			if (c != 'X') {
				cout<<"Malformed bound file: read "<<c
				    <<" instead of X"<<endl ;
				exit(-1) ;
			}
			// consume next lines about current task set
			for (int i = 1 ; i < N ; i++) {
				sim_res_file.
					ignore(numeric_limits
					       <streamsize>::max(), '\n');
				lat_bounds_file.
					ignore(numeric_limits
					       <streamsize>::max(), '\n');
			}
			sim_res_file.
				ignore(numeric_limits<streamsize>::max(), '\n');
			continue ;
		}
		DEB(cout<<"Read bound: "<<th_bound<<endl) ;

		for (int i = 0 ; i < N ; i++) {
			sim_res_file>>ts[i].avg_lat>>ts[i].max_lat ;
			
			lat_bounds_file>>th_bound ;

			DEB(cout<<"i = "<<i<<"/"<<N<<", read bound: "<<th_bound<<endl) ;
			
			if (sim_res_file.eof() || lat_bounds_file.eof()) {
				cout<<"Input files truncated for ts"
				    <<num_ts
				    <<endl ;
				exit(-1) ;
			}

			if (!lat_bounds_file) {
				cout<<"Malformed bound file"<<endl ;
				exit(-1) ;
			}
			
			double ratio = (th_bound - ts[i].max_lat) / ts[i].P ;
			
			ratios_file<<fixed<<ratio<<endl ;
		}
	}

	return 0 ;
}
