#!/usr/bin/env python
from __future__ import division
import schedcat.sched.edf.da as da
import schedcat.sched.edf.gel_pl as gel_pl

import schedcat.model.tasks as tasks

# Task set probably with millisecond parameters, but we
# will instead compute microsecond bounds.
from schedcat.util.time import ms2us

import sys

def read_next_ts(file, N):    
    for i in range(N):
        line = file.readline()
        if line == '':
            sys.exit()
        elements = line.strip().split(' ')
        #print float(elements[0]), float(elements[1]), float(elements[0]) / float(elements[1])
        yield tasks.SporadicTask(int(elements[0]), int(elements[1]))

def read_next_ts_lat(file, N):
    for i in range(N):
        line = file.readline()
        if line == '':
            sys.exit()
        elements = line.strip().split(' ')
        #print elements[1]
        yield float(elements[1])

if len(sys.argv) < 4 or sys.argv[1] == 'help':
    print 'Usage: compute_da_pl_bound.py <input filename> <bounds filename> <max-ratios filename>'
    print '       [da|pl] [reduction|m_2]'
    sys.exit()

# main
file = open(sys.argv[1])
bounds_file = open(sys.argv[2], 'w')
ratios_file = open(sys.argv[3], 'w')
line = file.readline()

old_M = 0

while line != '':
    elements = line.strip().split(' ')

    M = int(elements[0])

    if old_M == 0:
        old_M = M
    else:
        if old_M != M:
            print 'M mistmatch',old_M, M
            sys.exit(-1)
        
    N = int(elements[1]) + int(elements[2])

    #print M, N
    ts = tasks.TaskSystem(read_next_ts(file, N))

    actual_max_lats = read_next_ts_lat(file, N)

    if sys.argv[4] == 'pl':
        if len(sys.argv) == 6 and sys.argv[5] == 'm_2':
            success = gel_pl.bound_gedf_response_times(M, ts, 10, 1)
        else:
            success = gel_pl.bound_gedf_response_times(M, ts, 10, 0)
#            success = gel_pl.bound_gedf_response_times(M, ts, 0)
    else: # Devi-Anderson bound
        success = da.bound_response_times(M, ts)
        
    if success:
        lat_bounds = [task.response_time-task.deadline for task in ts]

        for bound in lat_bounds:
            bounds_file.write(str(bound))
            bounds_file.write('\n')

        ratios = [float(x)/y for x, y in zip(actual_max_lats, lat_bounds)]
        ratios_file.write('{:f}'.format(max(ratios)))
        ratios_file.write('\n')
    else:
        for actual_lat in actual_max_lats: # this also makes sure lats are read
            bounds_file.write('X')
            bounds_file.write('\n')

        ratios_file.write('X')
        ratios_file.write('\n')
        
    line = file.readline()

