#!/bin/bash
export LC_ALL=C
PLOT_TERM=$1 # Examples: aqua on OS X, x11 on Linux (if empty, no
	     # on-screen plot)
BASEDIR=${2:-test_results}
PROFILING=Y # set to Y to activate profiling (which deactivates
	    # parallel execution)
if [[ $PROFILING == "Y" ]]; then
    echo Profiling mode!!!!
fi
ONLY_HARM=no

echo Working in $BASEDIR

CALC_HARM_BOUND_RATIO=./calc-harm-bound_ratio-from-tasksets-BBP
#CALC_HARM_BOUND_RATIO=./calc-harm-bound_ratio-from-tasksets-BB
#CALC_HARM_BOUND_RATIO=./calc-harm-bound_ratio-from-tasksets-BF
#CALC_HARM_BOUND_RATIO=./calc-harm-M-2-bound_ratio-from-tasksets
CALC_NORM_ERROR=./calc-norm-error
CALC_AVG_AND_CO="./calc_avg_and_co.sh 95"

# MAC PATHS
COMPUTE_GEDF_DA_PL_BOUNDS="/Users/paolo/condivisa-MEGA/sviluppo_codice/real-time-dev/schedcat/compute_gedf-da-pl_bounds.py"

# Thinkpad paths
#COMPUTE_GEDF_DA_PL_BOUNDS="/home/paolo/schedcat/compute_gedf-da-pl_bounds.py"

NORM_CONF_INTERVALS=normalized_conf_intervals.txt
rm -f ${BASEDIR}/${NORM_CONF_INTERVALS}

function fill_min_avg_line {
    RATIOS_BNAME=$1
    type=$2
    i=$3
    RATIOS_STAT_BNAME=$4

    $CALC_AVG_AND_CO < ${RATIOS_BNAME}${type}-${i}CPU.txt \
		     > ${RATIOS_STAT_BNAME}${type}-${i}CPU.txt

    last_line=$(tail -n 1 ${RATIOS_STAT_BNAME}${type}-${i}CPU.txt)

    min=$(echo $last_line | awk '{printf " %f", $1}')

    avg=$(echo $last_line | awk '{printf " %f", $3}')

    conf=$(echo $last_line | awk '{printf " %f", $5}')

    if [[  "$min" != "" && "$avg" != "" && "$conf" != "" ]] ; then
	echo -n " $min" >> ${RATIOS_STAT_BNAME}min-table.txt
	echo -n " $avg" >> ${RATIOS_STAT_BNAME}avg-table.txt
	if [ 1 -eq "$(echo "$avg < 1000" | bc)" ] ; then
	    echo $(echo $conf/$avg | bc -l) >> \
		 ${BASEDIR}/${NORM_CONF_INTERVALS}
	    echo Normalized conf interval $(echo $conf/$avg | bc -l) for $avg
	fi
    else
	echo -n " X" >> ${RATIOS_STAT_BNAME}min-table.txt
	echo -n " X" >> ${RATIOS_STAT_BNAME}avg-table.txt
    fi
}

function make_profile_tables {
    VALUES_BNAME=$1
    PERIODS=$2
    UTOT=$3

    rm ${VALUES_BNAME}*table.txt
    TABLEFILE=${VALUES_BNAME}${PERIODS}-${UTOT}-table.txt
    for i in $RESULTS_TO_PARSE ; do
	echo $i
	echo -n $i >> $TABLEFILE

	for type in da reduced m_2 harm; do
	    if [[ -f ${VALUES_BNAME}${type}-${i}CPU.txt ]]; then
		echo -n " " >> $TABLEFILE
		cat ${VALUES_BNAME}${type}-${i}CPU.txt >> $TABLEFILE
	    else
		echo -n " X" >> ${VALUES_BNAME}table.txt
	    fi
	done

	echo >> $TABLEFILE
    done

    echo $(basename $TABLEFILE)
    cat $TABLEFILE
    echo
}

function make_stat_tables {
    VALUES_BNAME=$1
    VALUES_STAT_BNAME=$2

    rm -f ${VALUES_STAT_BNAME}min-table.txt ${VALUES_STAT_BNAME}avg-table.txt

    for i in $RESULTS_TO_PARSE ; do
	echo $i
	echo -n $i >> ${VALUES_STAT_BNAME}min-table.txt
	echo -n $i >> ${VALUES_STAT_BNAME}avg-table.txt

	for type in da reduced m_2 harm; do
	    echo Computing aggregate statistics with $type bound

	    echo fill_min_avg_line ${VALUES_BNAME} $type $i \
		 ${VALUES_STAT_BNAME}

	    fill_min_avg_line ${VALUES_BNAME} $type $i \
			      ${VALUES_STAT_BNAME}
	done

	echo >> ${VALUES_STAT_BNAME}min-table.txt
	echo >> ${VALUES_STAT_BNAME}avg-table.txt
    done

    echo $(basename ${VALUES_STAT_BNAME}min-table.txt)
    cat ${VALUES_STAT_BNAME}min-table.txt

    echo $(basename ${VALUES_STAT_BNAME}avg-table.txt)
    cat ${VALUES_STAT_BNAME}avg-table.txt
    echo
}

function make_plots {
    STAT_BNAME=$1
    LABEL="$2"
    MIN_Y=$3
    PLOT_TYPE=$4

    idx=$((${#STAT_BNAME}-1))
    LAST_CHAR=${STAT_BNAME:$idx:1}

    if [ $LAST_CHAR != "-" ]; then
	STAT_BNAME=${STAT_BNAME}-
    fi

    if [ "$PLOT_TERM" != "" ] ; then
	./make_plots.sh ${STAT_BNAME}table "$LABEL" $MIN_Y $PLOT_TERM $PLOT_TYPE
    fi

    ./make_plots.sh ${STAT_BNAME}table "$LABEL" $MIN_Y eps $PLOT_TYPE
    ./make_plots.sh ${STAT_BNAME}table "$LABEL" $MIN_Y gif $PLOT_TYPE
   if [[ $? -ne 0 ]]; then
	echo make_plots.sh failed
	exit
   fi
}

function execute_cmd {
    CMD="$1"
    i=$2
    type=$3
    multiplier=${4:-1}

    if [[ $PROFILING != "Y" ]]; then
	$CMD & # XXX CHECK WHETHER THE WAIT IN THE CALLER STILL WORKS !!!
    else
	echo $CMD
	TIMEFORMAT=%R
	TIME=$(time ($CMD) 2>&1 >comm_out.txt)

	if [ $(cat comm_out.txt | wc -l) -gt 1 ]; then
	    cat comm_out.txt
	    echo Command failed!
	    rm comm_out.txt
	    exit
	fi

	rm comm_out.txt
	echo Time: "$TIME * $multiplier"
	echo -n $(echo "$TIME * $multiplier" | bc) | \
	    tee $RES_PATH/${PROF_BASENAME}${type}-${i}CPU.txt
	echo " msec"
	echo Written to $RES_PATH/${PROF_BASENAME}${type}-${i}CPU.txt
    fi
}

function generate_stats {
    PERIODS=$1
    UTOT=$2
    TYPE=$3
    RES_PATH=$4/period-$PERIODS/$TYPE/ # This function works only if
				       # $4 contains a relative path
    RESULTS_TO_PARSE=${5:-"2 3 4 5 6 7 8"}
    MAX_NUM_TS=${6:-0}
    RESULTS_TO_MERGE=${7:-"6 7 8"}

    RES_BASENAME=sim_results-$TYPE-
    PARSED_BASENAME=parsed-$TYPE-
    NORM_ERROR_BASENAME=length_ratios-$TYPE-
    RATIOS_BASENAME=ratios-$TYPE-
    PROF_BASENAME=prof-$TYPE-
    BOUNDS_BASENAME=bounds-$TYPE-
    COMP_BASENAME=bound_comp-$PERIODS-${UTOT}-$TYPE-
    RATIOS_STAT_BASENAME=ratio_stats-$PERIODS-${UTOT}-$TYPE-
    NORM_ERROR_STAT_BASENAME=length_ratio_stats-$PERIODS-${UTOT}-$TYPE-

    # remove points from name, to avoid problems when using
    # final plots in latex documents
    RATIOS_STAT_BASENAME=$(echo $RATIOS_STAT_BASENAME | \
	sed 's/\.//g')
    NORM_ERROR_STAT_BASENAME=$(echo $NORM_ERROR_STAT_BASENAME | \
	sed 's/\.//g')

    echo Generating stats for $RES_PATH

#    if false; then # skipping stat computation

    for i in $RESULTS_TO_MERGE ; do
	if [ -f $RES_PATH/$PARSED_BASENAME${i}CPU.txt ]; then
	    continue
	fi
	for j in 1 2; do
	    echo $i $j
	    ./create_tasksets-from-medf-out.sh \
		$RES_PATH/$RES_BASENAME${i}CPU-${j}.txt > \
		$RES_PATH/$PARSED_BASENAME${i}CPU-${j}.txt
	    if [[ $? -ne 0 ]]; then
		    echo create_tasksets-from-medf-out.sh failed
		    exit
	    fi
	done

	cat $RES_PATH/$PARSED_BASENAME${i}CPU-2.txt \
	    >> $RES_PATH/$PARSED_BASENAME${i}CPU-1.txt

	mv $RES_PATH/$PARSED_BASENAME${i}CPU-1.txt \
	    $RES_PATH/$PARSED_BASENAME${i}CPU.txt

	rm $RES_PATH/$PARSED_BASENAME${i}CPU-2.txt
    done

    for i in $RESULTS_TO_PARSE ; do
	echo $i

	echo Generating parsed input files
	if [ -f $RES_PATH/$RES_BASENAME${i}CPU.txt ]; then
	    echo ./create_tasksets-from-medf-out.sh \
		$RES_PATH/$RES_BASENAME${i}CPU.txt > \
		$RES_PATH/$PARSED_BASENAME${i}CPU.txt
	    ./create_tasksets-from-medf-out.sh \
		$RES_PATH/$RES_BASENAME${i}CPU.txt > \
		$RES_PATH/$PARSED_BASENAME${i}CPU.txt
	    if [[ $? -ne 0 ]]; then
		    echo create_tasksets-from-medf-out.sh failed
		    exit
	    fi
	fi

	echo Computing harmonic bounds with $CALC_HARM_BOUND_RATIO ...

	# all the versions of CALC_HARM_BOUND_RATIO may become rather slow
	# if the number of processors is higher than 5. So reduce the number
	# of iterations to make simulations feasible
	if [[ $i -lt 6 ]]; then
	    max_ts=0
	else
	    if [[ $i -eq 6 ]]; then
		max_ts=100
	    else
		max_ts=10
	    fi
	fi
	max_ts=$(( $max_ts<$MAX_NUM_TS ? $max_ts : $MAX_NUM_TS ))

	if [[ $max_ts -gt 0 ]]; then
	    multiplier=$((1000 / $max_ts))
	else
	    multiplier=1
	fi

	HARM_CMD="$CALC_HARM_BOUND_RATIO \
	    $RES_PATH/$PARSED_BASENAME${i}CPU.txt \
	    $RES_PATH/${BOUNDS_BASENAME}harm-${i}CPU.txt \
	    $RES_PATH/${RATIOS_BASENAME}harm-${i}CPU.txt \
	    $RES_PATH/${PROF_BASENAME}harm_internal-${i}CPU.txt $max_ts"
	execute_cmd "$HARM_CMD" $i harm $multiplier

	if [ "$ONLY_HARM" == "yes" ]; then
		continue
	fi

	for type in da reduced m_2; do
	    echo Computing ${type} bounds ...
	    if [[ "$type" != "da" ]]; then
		supertype=pl
	    else
		supertype=
	    fi
	    CMD="$COMPUTE_GEDF_DA_PL_BOUNDS \
		$RES_PATH/$PARSED_BASENAME${i}CPU.txt \
		$RES_PATH/${BOUNDS_BASENAME}${type}-${i}CPU.txt \
		$RES_PATH/${RATIOS_BASENAME}${type}-${i}CPU.txt ${supertype} ${type}"
	    execute_cmd "$CMD" $i $type
	done

    done

    wait
#    fi # skipping stat computation

    if false; then # loop skipped, making only profiling tables
    for i in $RESULTS_TO_PARSE ; do
       echo $i
       for type in da reduced m_2 harm; do
# 	    echo Preparing data file on min ratios with $type bound
# 	    awk '{ if ($1 == "X") print "X"; else if ($1 > 0) printf "%f\n", 1.0/$1;\
# 			else print 1000000000 }' \
# 		< $RES_PATH/${RATIOS_BASENAME}${type}-${i}CPU.txt \
# 		> $RES_PATH/${RATIOS_BASENAME}inv-${type}-${i}CPU.txt &


	    echo Preparing data file on normalized errors with $type bound

	    echo $CALC_NORM_ERROR $RES_PATH/$PARSED_BASENAME${i}CPU.txt \
				$RES_PATH/${BOUNDS_BASENAME}${type}-${i}CPU.txt \
				${RES_PATH}/${NORM_ERROR_BASENAME}${type}-${i}CPU.txt

	    $CALC_NORM_ERROR $RES_PATH/$PARSED_BASENAME${i}CPU.txt \
				$RES_PATH/${BOUNDS_BASENAME}${type}-${i}CPU.txt \
				${RES_PATH}/${NORM_ERROR_BASENAME}${type}-${i}CPU.txt &

	done
    done
    wait
    fi # loop skipped, making only profiling tables

   make_profile_tables $RES_PATH/${PROF_BASENAME} ${PERIODS} ${UTOT}

   if false; then # make_stat_tables skipped, making only profiling tables
   make_stat_tables $RES_PATH/${RATIOS_BASENAME}inv- \
		    $RES_PATH/${RATIOS_STAT_BASENAME}inv- &

   make_stat_tables $RES_PATH/${NORM_ERROR_BASENAME} \
 		    $RES_PATH/${NORM_ERROR_STAT_BASENAME} &

   wait
   fi

#  for i in $RESULTS_TO_PARSE ; do
#	for type in da reduced m_2; do
# 	    echo Bound comparison against $type:
# 	    ./compare_bounds.py $RES_PATH/${BOUNDS_BASENAME}harm-${i}CPU.txt \
# 		$RES_PATH/${BOUNDS_BASENAME}${type}-${i}CPU.txt \
# 		> $RES_PATH/${COMP_BASENAME}${type}-${i}CPU.txt
#	done
#   done

#    tail -n 1 $RES_PATH/${RATIOS_STAT_BASENAME}inv-min-table.txt | \
# 	awk '{printf "Improvement DA-CVA2: %f DA-HARM: %f,  CVA2-HARM: %f\n", \
#             ($2-$4)/($2-1.0), ($2-$5)/($2-1.0), ($4-$5)/($4-1.0)}' > \
# 	$(echo $RES_PATH | cut -d "/" -f1)/${RATIOS_STAT_BASENAME}inv-improvement.txt
#    echo

   make_plots $RES_PATH/${PROF_BASENAME}${PERIODS}-${UTOT}- \
		"Average execution time [msec]" 0.01 exec # not executing in parallel
   						     # to not distort profiling

   return # skipping all the rest!!!!!!!

#    make_plots $RES_PATH/${RATIOS_STAT_BASENAME}inv-min "Minimum tightness index" 1 tard &
#
#    make_plots $RES_PATH/${RATIOS_STAT_BASENAME}inv-avg "Average tightness index" 1 tard &

    make_plots $RES_PATH/${NORM_ERROR_STAT_BASENAME}min \
 	       "Minimum normalized error" 0 tard &

    make_plots $RES_PATH/${NORM_ERROR_STAT_BASENAME}avg \
 	       "Average normalized error" 0 tard &

    wait
}

# Uniform distributions PROFILING
for U_pairs in 0_.99 .1_.2 .5_.99; do
       for period_pairs in 1-100 1-30 50-250; do
               	for Utot in 1; do # .5 .6 .7 .8 .9 1; do
			(generate_stats $period_pairs Utot_$Utot uniform-$U_pairs \
					${BASEDIR}/tot-utilization-${Utot}/uniform/\
					"2 3 4 5 6 7 8" 100)
		done
	done
done

# Bimodal distributions PROFILING
for heavy_distr in .272 .600 .705; do
    for period_pairs in 1-100 1-30 50-250; do
	for Utot in 1; do # .5 .6 .7 .8 .9 1; do
	    (generate_stats $period_pairs Utot_$Utot bimodal-$heavy_distr \
			    ${BASEDIR}/tot-utilization-${Utot}/bimodal/ \
			    "2 3 4 5 6 7 8" 100)
	done
    done
done

exit # EXIT EXIT EXIT EXIT EXIT

# Uniform distributions
for U_pairs in 0_.99 .1_.2 .5_.99; do
       for period_pairs in 1-100 1-30 50-250; do
               	for Utot in 1; do # .5 .6 .7 .8 .9 1; do
			(generate_stats $period_pairs Utot_$Utot uniform-$U_pairs \
    					${BASEDIR}/tot-utilization-${Utot}/uniform/ "2 3 4 5 6 7 8") &
		done
	done
done



# Bimodal distributions
for heavy_distr in .272 .600 .705; do
    for period_pairs in 1-100 1-30 50-250; do
	for Utot in 1; do # .5 .6 .7 .8 .9 1; do
	    (generate_stats $period_pairs Utot_$Utot bimodal-$heavy_distr \
			    ${BASEDIR}/tot-utilization-${Utot}/bimodal/) &
	done
    done
done

wait
$CALC_AVG_AND_CO < ${BASEDIR}/${NORM_CONF_INTERVALS}


# for U_pairs in .5_.99; do
#        for period_pairs in 50-250; do
#                	for Utot in .8 .9 1; do # .75 .8 .9 1; do
#  			(generate_stats $period_pairs Utot_${Utot} uniform-$U_pairs \
#     	${BASEDIR}/tot-utilization-${Utot}/uniform/) &
#  		done
#  	done
# done

# old profiling ....

#(generate_stats 1-100 full uniform-.5_.99 \
#    ${BASEDIR}/full-utilization/uniform/ "2")

#(generate_stats 1-100 full uniform-0_.99 \
#    ${BASEDIR}/full-utilization/uniform/ "8" 5)
