#!/bin/bash
LC_NUMERIC=C
file_name=$1 # without suffix
value_type=$2
min_y=${3:-1}
mode=${4:-"x11"}
type=${5:-"tard"}
plot_id=1

if [ $mode == "eps" ] ; then
	lw=3
else
	lw=1
fi


function write_basic_plot_conf()
{
    printf "
    set style data lp
    set mytics
    set xtics scale 0
    set grid y
    set ytics scale 2.0, 1.2
    set pointsize 2
    set key left samplen 1 width 2 height 0.5
    " >> $1
}

function plot_curves()
{
    topdir=$(echo $1 | cut -d "/" -f1)
    out_file_name=$(basename $1)
    out_file_name=$topdir/$out_file_name

    # make also a copy of the table
    cp $1.txt ${out_file_name}.txt

    x_label=${2:-"No x label!"}
    x_label_offset=${3:-0}
    min_y=$4
    y_label=$5
    plot_curves=$6
    ref_label=$7
    ref_value=$8
    plot_instr=${out_file_name}-gnuplo_comms.txt

    rm -f $plot_instr
    write_basic_plot_conf $plot_instr
    printf "
    set xlabel \"$x_label\" offset 0,$x_label_offset
    set ylabel \"$y_label\"
    #set yrange [$min_y:$max_y]
    " >> $plot_instr

    if [[ $type != speedup-4 && $type != speedup-8 ]]; then
	printf "
	       set logscale y
	       " >> $plot_instr
    fi

    if [[ $type == harm ]]; then
	printf "set yrange [$min_y:1000000]" >> $plot_instr
    else
	printf "set yrange [$min_y:]" >> $plot_instr
    fi

    case $mode in
	eps)
	    # WARNING: THIS TERMINAL SEEMS BROKEN IN GNUPLOT 5,
	    # WHICH ALWAYS USES THE SAME POINT TYPE FOR ALL CURVES
	printf "
        set output \"${out_file_name}.eps\"
        set term post eps 24
        " >> $plot_instr
	options="-mono"
	    ;;
	gif)
	printf "
        #set key horizontal 8000, 30
        set output \"${out_file_name}.gif\"
        set term gif 15
        " >> $plot_instr
	    ;;
	*)
	printf "
	set term $mode $plot_id font \"arial,12\"
        " >> $plot_instr
	options="-persist"
	    ;;
    esac

    printf "plot " >> $plot_instr

    if [ "$ref_label" != "" ] ; then
	printf "%f t \"$ref_label\" lw $lw, " $ref_value >> $plot_instr
    fi

    echo $plot_curves >> $plot_instr

    gnuplot $options < $plot_instr
    rm -f $plot_instr

    plot_id=$(($plot_id+1))

    if [[ $mode != "x11" && $mode != "aqua" ]] ; then
    	echo Wrote $out_file_name.$mode
    fi
}

function plot_tardiness()
{
    file_name=$1
    y_label="$2"
    min_y=$3

    plot_curves $file_name \
	"Number of processors" 0 $min_y "$y_label" \
	"\"$file_name.txt\" \
             using 2:xtic(1) t \"DA\", \
             \"\" using 3 t \"CVA\", \
             \"\" using 4 t \"CVA2\", \
             \"\" using 5 t \"HARM\" "
#	"Ideal value: 1" 1
}

function plot_exec_time()
{
    file_name=$1
    y_label="$2"
    min_y=$3

    plot_curves $file_name \
	"Number of processors" 0 $min_y "$y_label" \
	"\"$file_name.txt\" \
             using 2:xtic(1) t \"da\", \
             \"\" using 3 t \"cva\", \
             \"\" using 4 t \"cva2\", \
	     \"\" using 5 t \"harm-BB-4\" "
#	"Ideal value: 1" 1
}

function plot_exec_time_harm()
{
    file_name=$1
    y_label="$2"
    min_y=$3

    # max_y will be set automatically in plot_curves!!!!
    echo invoking plot_curves

    plot_curves $file_name \
	"Number of processors" 0 $min_y "$y_label" \
	"\"$file_name.txt\" \
             using 2:xtic(1) t \"harm-BF\", \
	\"\" using 3 t \"harm-BB-4\" "
}

function plot_exec_time_harm_1_4()
{
    file_name=$1
    y_label="$2"
    min_y=$3

    plot_curves $file_name \
	"Number of processors" 0 $min_y "$y_label" \
	"\"$file_name.txt\" \
             using 2:xtic(1) t \"harm-BB-1\", \
	     \"\" using 3 t \"harm-BB-4\" "
}

function plot_speedup()
{
    file_name=$1
    y_label="$2"
    min_y=$3

    plot_curves $file_name \
	"Number of parallel threads" 0 $min_y "$y_label" \
	"\"$file_name.txt\" \
             using 2:xtic(1) t \"harm-BB, 8-processor case\""
}


case $type in
    tard)
	echo plot_tardiness $file_name "$value_type" $min_y
	plot_tardiness $file_name "$value_type" $min_y
	;;
    harm)
	plot_exec_time_harm $file_name "$value_type" $min_y
	;;
    exec-times-4)
	plot_exec_time_harm_1_4 $file_name "$value_type" $min_y
	;;
    speedup-4)
	plot_speedup $file_name "Speedup on Dual-Core CPU" $min_y
	;;
    speedup-8)
	plot_speedup $file_name "Speedup on Quad-Core CPU" $min_y
	;;
    *)
	plot_exec_time $file_name "$value_type" $min_y
	;;
esac
