#!/bin/bash
INITIAL_SEED=${1:-1} # DO NOT use 0 as initial seed!!!
cur_seed=$INITIAL_SEED
min_M=2
max_M=6

function do_simulations {
	params="$1"
	seed=$2
	outfile=$3

	for ((i = $min_M ; i <= $max_M ; i++)); do
		echo $(( $i + $seed )) > rand_gen.txt
		echo Seed: $(cat rand_gen.txt)
		if [[ $i -gt 5 ]]; then
			echo "./medf $i $i 500 $params > $outfile${i}CPU-1.txt &"
			./medf $i $i 500 $params > $outfile${i}CPU-1.txt &
			sleep 1

			echo $(( $i + ( $seed * 10 ) )) > rand_gen.txt
			echo Seed: $(cat rand_gen.txt)
			echo "./medf $i $i 500 $params > $outfile${i}CPU-2.txt &"
			./medf $i $i 500 $params > $outfile${i}CPU-2.txt &
		else
			echo "./medf $i $i 1000 $params > $outfile${i}CPU.txt &"
			./medf $i $i 1000 $params > $outfile${i}CPU.txt &
		fi
		sleep 1
	done
}

function sim-uniform-distribution {
	max_period=$1
	min_period=$2
	min_U=$3
	max_U=$4
	Utot=${5:-1}

	mkdir -p results/tot-utilization-$Utot/uniform/period-$min_period-$max_period
	BASE_DIR=results/tot-utilization-$Utot/uniform/period-$min_period-$max_period/uniform-${min_U}_${max_U}

	mkdir -p $BASE_DIR
	#rm -rf $BASE_DIR/*
	do_simulations "$min_U $max_U 1 $Utot $max_period $min_period" $cur_seed $BASE_DIR/sim_results-uniform-${min_U}_${max_U}-
}

function sim-bimodal-distribution {
	max_period=$1
	min_period=$2
	heavy_prob=$3 # probability of utilizations in [0.5, 0.99]
	Utot=${4:-1}

	# ratio between heavy-task and light-task fractions
	fractions_ratio=$( echo "(.75 * ($heavy_prob) )/(.25 * ( 1-($heavy_prob) ) )" | bc -l)
	
	heavy_fraction=$(echo "scale = 3; $fractions_ratio / ($fractions_ratio + 1)" | bc -l)

	BASE_DIR=results/tot-utilization-$Utot/bimodal/period-$min_period-$max_period/bimodal-$heavy_fraction
	mkdir -p $BASE_DIR
	#rm -rf $BASE_DIR/*

	do_simulations "0.5 .99 $heavy_fraction $Utot $max_period $min_period" $cur_seed $BASE_DIR/sim_results-bimodal-$heavy_fraction-
}

# Uniform distributions
for U_pairs in ".1 .2" "0 .99" ".5 .99"; do
	for period_pairs in "100 1" "30 1" "250 50"; do
		for Utot in .5 .6 .7 .8 .9 1; do
			sim-uniform-distribution $period_pairs $U_pairs $Utot
		done
	done
done
 
# Bimodal distributions
for heavy_distr in "1/9" "3/9" "4/9"; do
	for period_pairs in "100 1" "30 1" "250 50"; do
		for Utot in .5 .6 .7 .8 .9 1; do
			sim-bimodal-distribution $period_pairs $heavy_distr $Utot
		done
	done
done

